use palette::Srgb;
use serde::{Deserialize, Serialize};

use crate::dither::DitheredImage;

/// Palette used on the display; pixels can be one of these colors.
///
/// The RGB values are slightly adjusted to improve accuracy.
const DISPLAY_PALETTE: [Srgb; 7] = [
    Srgb::new(0.047, 0.047, 0.055), // Black
    Srgb::new(0.824, 0.824, 0.816), // White
    Srgb::new(0.118, 0.376, 0.122), // Green
    Srgb::new(0.114, 0.118, 0.667), // Blue
    Srgb::new(0.549, 0.106, 0.114), // Red
    Srgb::new(0.827, 0.788, 0.239), // Yellow
    Srgb::new(0.757, 0.443, 0.165), // Orange
];



/// A more primitive palette that doesn't reflect reality. Here for posterity.
/// This is based on the datasheet, not on empirical testing
const SIMPLE_PALETTE: [Srgb; 7] = [
    Srgb::new(0.0, 0.0, 0.0),       // Black
    Srgb::new(1.0, 1.0, 1.0),       // White
    Srgb::new(0.0, 1.0, 0.0),       // Green
    Srgb::new(0.0, 0.0, 1.0),       // Blue
    Srgb::new(1.0, 0.0, 0.0),       // Red
    Srgb::new(1.0, 1.0, 0.0),       // Yellow
    Srgb::new(0.757, 0.443, 0.165), // Orange
];

#[derive(strum::EnumString, strum::Display, Serialize, Deserialize, PartialEq, Eq, Debug, Clone, Copy)]
pub enum Palette {
    Default,
    Simple,
}

impl Palette {
    #[must_use]
    pub const fn value(&self) -> &[Srgb] {
        match self {
            Self::Default => &DISPLAY_PALETTE,
            Self::Simple => &SIMPLE_PALETTE,
        }
    }
}



/// Construct a dithered image for the EINK display using the default palette and correct
/// resolution.
#[must_use]
pub fn new_image() -> DitheredImage {
    DitheredImage::new(800,480, DISPLAY_PALETTE.to_vec())
}
