pub mod api;
pub mod display;
pub mod dither;
pub mod eink;

use serde::Deserialize;
use std::path::PathBuf;
use toml;

use crate::display::get_display;
use crate::dither::{DitherMethod, DitheredImage};
use crate::eink::Palette;
use clap::{Args, Parser, Subcommand};
use image::RgbImage;
use tracing::{error, event, info, instrument, warn, Level};

/// Application config, including sqlite db path, scan folders, and scheduling.
#[derive(Deserialize, Debug)]
struct Config {
    database_path: PathBuf,
    search_paths: Vec<PathBuf>,
}

/// Display images on E-Ink Displays
#[derive(Debug, Parser)]
#[command(version, about)]
struct Cli {
    #[command(subcommand)]
    command: Command,
}

#[derive(Debug, Subcommand)]
enum Command {
    /// Convert an image into a dithered form.
    Convert(ConvertArgs),
    /// Load a single image
    Show,
    /// Start the HTTP server
    Serve,
}

#[derive(Args, Debug)]
struct ConvertArgs {
    #[arg(short, long, default_value_t = DitherMethod::NearestNeighbor)]
    dither_method: DitherMethod,
    #[arg(short, long, default_value_t = Palette::Default)]
    palette: Palette,
    input_file: PathBuf,
    output_file: PathBuf,
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    tracing_subscriber::fmt::init();
    let cli = Cli::parse();

    match cli.command {
        Command::Convert(a) => {
            let input = image::ImageReader::open(a.input_file)?
                .decode()?
                .into_rgb8();
            let mut method = a.dither_method.get_ditherer();

            let mut result = DitheredImage::new(
                input.width(),
                input.height(),
                Palette::Default.value().to_vec(),
            );
            method.dither(&input, &mut result);

            result.into_rgbimage().save(a.output_file)?;
        }
        Command::Show => {
            let mut display = get_display();
            let img: RgbImage = image::ImageReader::open("image.png")?.decode()?.into();
            error!("HI");

            let mut eink_buf = crate::eink::new_image();
            let mut dither = DitherMethod::Atkinson.get_ditherer();

            dither.dither(&img, &mut eink_buf);
            display.display(&eink_buf)?;
        }
        Command::Serve => {
            let display = get_display();
            let ctx = api::AppState::new(display);

            let app = api::router().with_state(ctx);
            let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await?;
            info!("Listening on 0.0.0.0:3000");
            axum::serve(listener, app).await?;
        }
    }

    Ok(())
}
